from typing import cast

import numpy as np
import pandas as pd
from numpy.testing import assert_allclose, assert_array_equal
from numpy.typing import NDArray
from pandas._testing.asserters import assert_series_equal
from pandas.testing import assert_frame_equal

from saiph.reduction import DUMMIES_PREFIX_SEP
from saiph.reduction.famd_sparse import (
    center_sparse,
    fit,
    fit_transform,
    scaler_sparse,
    transform,
)
from saiph.reduction.pca import center as center_pca
from saiph.reduction.pca import fit_transform as fit_pca
from saiph.reduction.pca import scaler as scaler_pca


def test_fit_mix(mixed_df2: pd.DataFrame) -> None:
    df = mixed_df2.iloc[:, 0:3]

    result, model = fit_transform(df)
    model.mean = cast(pd.Series, model.mean)
    model.std = cast(pd.Series, model.std)
    model.prop = cast(pd.Series, model.prop)
    model.s = cast(NDArray[np.float_], model.s)

    expected_result = pd.DataFrame(
        {
            "Dim. 1": [-1.73, 1.73],
            "Dim. 2": [1.41, 1.41],
        }
    )
    expected_v: NDArray[np.float_] = np.array(
        [
            [-5.773503e-01, -4.082483e-01, 4.082483e-01, 4.082483e-01, -4.082483e-01],
            [5.384773e-16, 5.000000e-01, 5.000000e-01, 5.000000e-01, 5.000000e-01],
        ]
    )
    expected_s: NDArray[np.float_] = np.array([1.224745e00, 1])
    expected_u: NDArray[np.float_] = np.array([[1.0, 1.0], [-1.0, 1.0]])
    expected_explained_var: NDArray[np.float_] = np.array([1.5, 1.0])
    expected_explained_var_ratio: NDArray[np.float_] = np.array([0.6, 0.4])

    assert_frame_equal(abs(result), abs(expected_result), check_exact=False, atol=0.01)
    assert_allclose(np.abs(model.V), np.abs(expected_v), atol=0.01)
    assert_allclose(model.s, expected_s, atol=0.01)
    assert_allclose(np.abs(model.U), np.abs(expected_u), atol=0.01)
    assert_allclose(model.explained_var, expected_explained_var, atol=0.01)
    assert_allclose(model.explained_var_ratio, expected_explained_var_ratio, atol=0.01),
    assert_allclose(model.variable_coord, model.V.T)
    assert np.array_equal(
        model._modalities,  # type: ignore
        [
            f"tool{DUMMIES_PREFIX_SEP}hammer",
            f"tool{DUMMIES_PREFIX_SEP}toaster",
            f"score{DUMMIES_PREFIX_SEP}aa",
            f"score{DUMMIES_PREFIX_SEP}ab",
        ],
    )
    assert model.D_c is None

    assert_allclose(model.mean, np.array(2.5))
    assert_allclose(model.std, np.array(1.5))

    assert_allclose(
        model.prop,
        [0.5, 0.5, 0.5, 0.5],
        atol=0.01,
    )
    assert np.array_equal(
        model._modalities,  # type: ignore
        [
            f"tool{DUMMIES_PREFIX_SEP}hammer",
            f"tool{DUMMIES_PREFIX_SEP}toaster",
            f"score{DUMMIES_PREFIX_SEP}aa",
            f"score{DUMMIES_PREFIX_SEP}ab",
        ],
    )


def test_transform(mixed_df2: pd.DataFrame) -> None:
    df = mixed_df2

    model = fit(df)

    df_transformed = transform(df, model)
    df_expected = pd.DataFrame(
        {
            "Dim. 1": [2.0, -2],
            "Dim. 2": [1.414214, 1.414214],
        }
    )
    assert_frame_equal(df_transformed.abs(), df_expected.abs())


def test_transform_vs_coord(mixed_df2: pd.DataFrame) -> None:
    df = mixed_df2

    coord, model = fit_transform(df)
    df_transformed = transform(df, model)

    assert_frame_equal(df_transformed, coord)


def test_fit_zero() -> None:
    df = pd.DataFrame(
        {
            "tool": ["toaster", "toaster"],
            "score": ["aa", "aa"],
        }
    )
    result, _ = fit_transform(df)

    expected = pd.DataFrame(
        {
            "Dim. 1": [1.414213562373095, 1.414213562373095],
            "Dim. 2": [-1.110223e-16, -1.110223e-16],
        }
    )
    assert_frame_equal(result, expected, check_exact=False, atol=0.01)


def test_scaler_pca_famd(mixed_df2: pd.DataFrame) -> None:
    """Compare FAMD and PCA numeric scaler results."""
    original_df = mixed_df2

    _, model = fit_transform(original_df)
    df_famd = scaler_sparse(model, original_df)

    _, model_pca = fit_pca(original_df[model.original_continuous])
    df_pca = scaler_pca(model_pca, original_df)

    assert_array_equal(
        df_famd.todense()[:, [0, 1]], df_pca[model.original_continuous].to_numpy()
    )


def test_center_pca_famd(mixed_df2: pd.DataFrame) -> None:
    """Compare FAMD and PCA numeric center results."""
    original_df = mixed_df2

    _, model = fit_transform(original_df)
    continuous = model.original_continuous
    categorical = model.original_categorical
    df_famd, mean1, std1, _, _ = center_sparse(
        original_df, quali=categorical, quanti=continuous
    )

    df_pca, mean2, std2 = center_pca(original_df[continuous])

    assert_array_equal(df_famd.todense()[:, [0, 1]], df_pca.to_numpy())

    assert_series_equal(mean1, mean2)
    assert_series_equal(std1, std2)



"""PCA projection module."""
import sys
from typing import Optional, Tuple

import numpy as np
import pandas as pd
from numpy.typing import NDArray

from saiph.models import Model
from saiph.reduction.utils.common import (
    get_explained_variance,
    get_projected_column_names,
    get_uniform_row_weights,
)
from saiph.reduction.utils.svd import get_svd


def fit(
    df: pd.DataFrame,
    nf: Optional[int] = None,
    col_weights: Optional[NDArray[np.float_]] = None,
    seed: Optional[int] = None,
) -> Model:
    """Fit a PCA model on data.

    Parameters:
        df: Data to project.
        nf: Number of components to keep. default: min(df.shape)
        col_weights: Weight assigned to each variable in the projection
            (more weight = more importance in the axes). default: np.ones(df.shape[1])

    Returns:
        model: The model for transforming new data.
    """
    nf = nf or min(df.shape)
    _col_weights = col_weights if col_weights is not None else np.ones(df.shape[1])

    # Set row weights
    row_w = get_uniform_row_weights(len(df))

    df_centered, mean, std = center(df)

    # Apply weights and compute svd
    Z = ((df_centered * _col_weights).T * row_w).T
    U, S, Vt = get_svd(Z, nf=nf, seed=seed)

    U = ((U.T) / np.sqrt(row_w)).T
    Vt = Vt / np.sqrt(_col_weights)

    explained_var, explained_var_ratio = get_explained_variance(S, df.shape[0], nf)

    # Retain only the nf higher singular values
    U = U[:, :nf]
    S = S[:nf]
    Vt = Vt[:nf, :]

    model = Model(
        original_dtypes=df.dtypes,
        original_categorical=[],
        original_continuous=df.columns.to_list(),
        dummy_categorical=[],
        U=U,
        V=Vt,
        explained_var=explained_var,
        explained_var_ratio=explained_var_ratio,
        variable_coord=pd.DataFrame(Vt.T),
        mean=mean,
        std=std,
        type="pca",
        is_fitted=True,
        nf=nf,
        column_weights=_col_weights,
        row_weights=row_w,
        modalities_types={},
    )

    return model


def fit_transform(
    df: pd.DataFrame,
    nf: Optional[int] = None,
    col_weights: Optional[NDArray[np.float_]] = None,
    seed: Optional[int] = None,
) -> Tuple[pd.DataFrame, Model]:
    """Fit a PCA model on data and return transformed data.

    Parameters:
        df: Data to project.
        nf: Number of components to keep. default: min(df.shape)
        col_weights: Weight assigned to each variable in the projection
            (more weight = more importance in the axes). default: np.ones(df.shape[1])

    Returns:
        model: The model for transforming new data.
        coord: The transformed data.
    """
    model = fit(df, nf, col_weights, seed=seed)
    coord = transform(df, model)
    return coord, model


def center(df: pd.DataFrame) -> Tuple[pd.DataFrame, pd.Series, pd.Series]:
    """Center data and standardize it if scale. Compute mean and std values.

    Used as internal function during fit.

    **NB**: saiph.reduction.pca.scaler is better suited when a Model is already fitted.

    Parameters:
        df: DataFrame to center.

    Returns:
        df: The centered DataFrame.
        mean: Mean of the input dataframe.
        std: Standard deviation of the input dataframe.
    """
    df = df.copy()
    mean = np.mean(df, axis=0)
    df -= mean

    std = np.std(df, axis=0)
    std[std <= sys.float_info.min] = 1
    df /= std

    return df, mean, std


def scaler(model: Model, df: pd.DataFrame) -> pd.DataFrame:
    """Scale data using mean and std from model.

    Parameters:
        model: Model computed by fit.
        df: DataFrame to scale.

    Returns:
        df: The scaled DataFrame.
    """
    df_scaled = df.copy()

    df_scaled -= model.mean
    df_scaled /= model.std
    return df_scaled


def transform(df: pd.DataFrame, model: Model) -> pd.DataFrame:
    """Scale and project into the fitted numerical space.

    Parameters:
        df: DataFrame to transform.
        model: Model computed by fit.

    Returns:
        coord: Coordinates of the dataframe in the fitted space.
    """
    df_scaled = scaler(model, df)
    coord = df_scaled @ model.V.T
    coord.columns = get_projected_column_names(model.nf)
    return coord



# type: ignore
import time
import webbrowser
from pathlib import Path

import pandas as pd
import typer
from filprofiler.api import profile

from saiph import fit
from saiph.lib.size import get_readable_size

N_ROWS = 10000

# Run with:
# poetry run fil-profile python saiph/tests/profile_memory.py True

BASE_PATH = (Path(__file__).parent / "../../").resolve()


def main(sparse: bool) -> None:
    """Profile famd.fit using a fake dataset."""
    df = pd.read_csv(str(BASE_PATH / "tmp/fake_1000000.csv"))

    typer.echo(f"using {get_readable_size(df.memory_usage(index=True).sum())}")

    typer.echo("before fit")
    start = time.perf_counter()
    filename = f"tmp/{time.time()}"
    full_path = BASE_PATH / filename / "index.html"
    profile(lambda: fit(df, nf=5, sparse=sparse), filename)
    end = time.perf_counter()

    typer.echo(f"after fit, took {(end-start):.3} sec")

    webbrowser.open(f"file://{str(full_path)}")


if __name__ == "__main__":
    typer.run(main)



import numpy as np
import pandas as pd
import pytest

from saiph.contribution import filter_variable_contribution


@pytest.fixture
def contributions() -> pd.DataFrame:
    df = pd.DataFrame(
        data=[[1, 0.01], [2, 1]],
        index=["Var. 1", "Var. 2"],
        columns=["Dim. 1", "Dim. 2"],
    )

    return df


def test_filter_variable_contribution_is_sorted(contributions: pd.DataFrame) -> None:
    """Verify that we get sorted descending values."""
    result = filter_variable_contribution(
        contributions=contributions, dim=2, max_var=None, min_contrib=None
    )
    np.testing.assert_array_equal(result.values, [1, 0.01])
    np.testing.assert_array_equal(result.index.values, ["Var. 2", "Var. 1"])


def test_filter_variable_contribution_filter_by_min_contrib(
    contributions: pd.DataFrame,
) -> None:
    """Verify that we can filter by minimum contribution."""
    result = filter_variable_contribution(
        contributions=contributions, dim=2, max_var=None, min_contrib=0.5
    )
    np.testing.assert_array_equal(result.values, [1])


def test_filter_variable_contribution_filter_by_max_var(
    contributions: pd.DataFrame,
) -> None:
    """Verify that we can filter by maximum number of variables."""
    result = filter_variable_contribution(contributions=contributions, dim=2, max_var=1)
    np.testing.assert_array_equal(result.values, [1])



"""FAMD projection module."""
import sys
from itertools import chain, repeat
from typing import Any, Callable, List, Optional, Tuple, cast

import numpy as np
import pandas as pd
import scipy
from numpy.typing import NDArray

from saiph.models import Model
from saiph.reduction import DUMMIES_PREFIX_SEP
from saiph.reduction.utils.common import (
    column_multiplication,
    get_dummies_mapping,
    get_explained_variance,
    get_grouped_modality_values,
    get_modalities_types,
    get_projected_column_names,
    get_uniform_row_weights,
    row_division,
    row_multiplication,
)
from saiph.reduction.utils.svd import get_svd


def center(
    df: pd.DataFrame, quanti: List[str], quali: List[str]
) -> Tuple[
    pd.DataFrame, NDArray[np.float_], NDArray[np.float_], NDArray[Any], NDArray[Any]
]:
    """Center data, scale it, compute modalities and proportions of each categorical.

    Used as internal function during fit.

    **NB**: saiph.reduction.famd.scaler is better suited when a Model is already fitted.

    Parameters:
        df: DataFrame to center.
        quanti: Indices of continuous variables.
        quali: Indices of categorical variables.

    Returns:
        df_scale: The scaled DataFrame.
        mean: Mean of the input dataframe.
        std: Standard deviation of the input dataframe.
        prop: Proportion of each categorical.
        _modalities: Modalities for the MCA.
    """
    # Scale the continuous data
    df_quanti = df[quanti]
    mean = np.mean(df_quanti, axis=0)
    df_quanti -= mean
    std = np.std(df_quanti, axis=0)
    std[std <= sys.float_info.min] = 1
    df_quanti /= std

    # scale the categorical data
    df_quali = pd.get_dummies(
        df[quali].astype("category"), prefix_sep=DUMMIES_PREFIX_SEP
    )
    # .mean() is the same as counting 0s and 1s
    # This will only work if we stick with pd.get_dummies to encode the modalities
    # and we don't use `drop_first=True`
    prop = df_quali.mean()
    df_quali -= prop
    df_quali /= np.sqrt(prop)
    _modalities = df_quali.columns.values

    df_scale = pd.concat([df_quanti, df_quali], axis=1)

    return df_scale, mean, std, prop, _modalities


def fit(
    df: pd.DataFrame,
    nf: Optional[int] = None,
    col_weights: Optional[NDArray[np.float_]] = None,
    center: Callable[
        [pd.DataFrame, List[str], List[str]],
        Tuple[
            pd.DataFrame,
            NDArray[np.float_],
            NDArray[np.float_],
            NDArray[Any],
            NDArray[Any],
        ],
    ] = center,
    seed: Optional[int] = None,
) -> Model:
    """Fit a FAMD model on data.

    Parameters:
        df: Data to project.
        nf: Number of components to keep. default: min(df.shape)
        col_weights: Weight assigned to each variable in the projection
            (more weight = more importance in the axes). default: np.ones(df.shape[1])

    Returns:
        model: The model for transforming new data.
    """
    nf = nf or min(pd.get_dummies(df).shape)
    _col_weights = np.ones(df.shape[1]) if col_weights is None else col_weights

    # Select the categorical and continuous columns
    quanti = df.select_dtypes(include=["int", "float", "number"]).columns.to_list()
    quali = df.select_dtypes(exclude=["int", "float", "number"]).columns.to_list()
    dummy_categorical = pd.get_dummies(
        df[quali].astype("category"), prefix_sep=DUMMIES_PREFIX_SEP
    ).columns.to_list()
    modalities_types = get_modalities_types(df[quali])

    row_w = get_uniform_row_weights(len(df))
    col_weights = _col_weights_compute(df, _col_weights, quanti, quali)

    df_scaled, mean, std, prop, _modalities = center(df, quanti, quali)

    # Apply the weights
    Z = df_scaled.multiply(col_weights).T.multiply(row_w).T

    # Compute the svd
    _U, S, _Vt = (
        get_svd(Z.todense(), nf=nf, seed=seed)
        if isinstance(Z, scipy.sparse.spmatrix)
        else get_svd(Z, nf=nf, seed=seed)
    )

    U = ((_U.T) / np.sqrt(row_w)).T
    Vt = _Vt / np.sqrt(col_weights)

    explained_var, explained_var_ratio = get_explained_variance(S, df.shape[0], nf)

    # Retain only the nf higher singular values
    U = U[:, :nf]
    S = S[:nf]
    Vt = Vt[:nf, :]

    model = Model(
        original_dtypes=df.dtypes,
        original_categorical=quali,
        original_continuous=quanti,
        dummy_categorical=dummy_categorical,
        U=U,
        V=Vt,
        s=S,
        explained_var=explained_var,
        explained_var_ratio=explained_var_ratio,
        variable_coord=pd.DataFrame(Vt.T),
        mean=mean,
        std=std,
        prop=prop,
        _modalities=_modalities,
        type="famd",
        is_fitted=True,
        nf=nf,
        column_weights=col_weights,
        row_weights=row_w,
        modalities_types=modalities_types,
    )

    return model


def fit_transform(
    df: pd.DataFrame,
    nf: Optional[int] = None,
    col_weights: Optional[NDArray[np.float_]] = None,
    seed: Optional[int] = None,
) -> Tuple[pd.DataFrame, Model]:
    """Fit a FAMD model on data and return transformed data.

    Parameters:
        df: Data to project.
        nf: Number of components to keep. default: min(df.shape)
        col_weights: Weight assigned to each variable in the projection
            (more weight = more importance in the axes). default: np.ones(df.shape[1])

    Returns:
        coord: The transformed data.
        model: The model for transforming new data.
    """
    model = fit(df, nf, col_weights, seed=seed)
    coord = transform(df, model)
    return coord, model


def _col_weights_compute(
    df: pd.DataFrame, col_weights: NDArray[Any], quanti: List[int], quali: List[int]
) -> NDArray[Any]:
    """Calculate weights for columns given what weights the user gave."""
    # Set the columns and row weights
    weight_df = pd.DataFrame([col_weights], columns=df.columns)
    weight_quanti = weight_df[quanti]
    weight_quali = weight_df[quali]

    # Get the number of modality for each quali variable
    modality_numbers = []
    for column in weight_quali.columns:
        modality_numbers += [len(df[column].unique())]

    # Set weight vector for categorical columns
    weight_quali_rep = list(
        chain.from_iterable(
            repeat(i, j) for i, j in zip(list(weight_quali.iloc[0]), modality_numbers)
        )
    )

    _col_weights: NDArray[Any] = np.array(
        list(weight_quanti.iloc[0]) + weight_quali_rep
    )

    return _col_weights


def scaler(model: Model, df: pd.DataFrame) -> pd.DataFrame:
    """Scale data using mean, std, modalities and proportions of each categorical from model.

    Parameters:
        model: Model computed by fit.
        df: DataFrame to scale.

    Returns:
        df_scaled: The scaled DataFrame.
    """
    model.prop = cast(pd.Series, model.prop)

    df_quanti = df[model.original_continuous]
    df_quanti = (df_quanti - model.mean) / model.std

    # scale
    df_quali = pd.get_dummies(
        df[model.original_categorical].astype("category"), prefix_sep=DUMMIES_PREFIX_SEP
    )
    if model._modalities is not None:
        for mod in model._modalities:
            if mod not in df_quali:
                df_quali[mod] = 0
    df_quali = df_quali[model._modalities]
    df_quali = (df_quali - model.prop) / np.sqrt(model.prop)

    df_scaled = pd.concat([df_quanti, df_quali], axis=1)
    return df_scaled


def transform(
    df: pd.DataFrame,
    model: Model,
    *,
    scaler: Callable[[Model, pd.DataFrame], pd.DataFrame] = scaler,
) -> pd.DataFrame:
    """Scale and project into the fitted numerical space.

    Parameters:
        df: DataFrame to transform.
        model: Model computed by fit.

    Returns:
        coord: Coordinates of the dataframe in the fitted space.
    """
    df_scaled = scaler(model, df)
    coord = pd.DataFrame(df_scaled @ model.V.T)

    coord.columns = get_projected_column_names(len(coord.columns))
    return coord


def stats(model: Model, df: pd.DataFrame, explode: bool = False) -> Model:
    """Compute contributions and cos2.

    Parameters:
        model: Model computed by fit.
        df: dataframe to compute statistics from
        explode: whether to split the contributions of each modality (True)
            or sum them as the contribution of the whole variable (False)

    Returns:
        model: model populated with contribution and cos2.
    """
    if not model.is_fitted:
        raise ValueError(
            "Model has not been fitted. Call fit() to create a Model instance."
        )

    contributions, cos2 = get_variable_contributions(model, df, explode=explode)
    model.contributions = contributions
    model.cos2 = cos2
    return model


def get_variable_contributions(
    model: Model, df: pd.DataFrame, explode: bool = False
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Compute the contributions of the `df` variables within the fitted space.

    Parameters:
        model: Model computed by fit.
        df: dataframe to compute contributions from
        explode: whether to split the contributions of each modality (True)
            or sum them as the contribution of the whole variable (False)

    Returns:
        tuple of contributions and cos2.
    """
    scaled_df = pd.DataFrame(scaler(model, df))
    weighted_df = column_multiplication(scaled_df, np.sqrt(model.column_weights))
    weighted_df = row_multiplication(weighted_df, np.sqrt(model.row_weights))

    min_nf = min(min(weighted_df.shape), model.nf)
    s, U, eig = _compute_svd(model, weighted_df, min_nf=min_nf)
    contributions = _compute_contributions(
        model, s, U, eig, min_nf, scaled_df.columns, explode=explode
    )

    continuous_cos2 = compute_continuous_cos2(model, scaled_df, min_nf, s, U)
    categorical_cos2 = compute_categorical_cos2(model, df, min_nf)
    combined_cos2 = pd.concat([continuous_cos2, categorical_cos2])

    return contributions, combined_cos2


def _compute_contributions(
    model: Model,
    s: NDArray[np.float_],
    U: NDArray[np.float_],
    eig: NDArray[np.float_],
    min_nf: int,
    column_names: List[str],
    *,
    explode: bool = True,
) -> pd.DataFrame:
    # compute the contribution
    raw_contributions = (U * s) ** 2 / eig
    raw_contributions = raw_contributions * model.column_weights[:, np.newaxis]
    summed_contributions: NDArray[np.float_] = np.array(raw_contributions.sum(axis=0))
    raw_contributions /= summed_contributions
    raw_contributions *= 100

    contributions = pd.DataFrame(
        raw_contributions,
        index=column_names,
        columns=get_projected_column_names(min_nf),
    )

    if explode:
        return contributions

    mapping = get_dummies_mapping(model.original_categorical, model.dummy_categorical)

    contributions = get_grouped_modality_values(mapping, contributions)

    return contributions


def compute_categorical_cos2(
    model: Model, df: pd.DataFrame, min_nf: int
) -> pd.DataFrame:
    """Compute the cos2 statistic for categorical variables.

    Parameters
    ----------
    model :
        model
    df :
        dataframe
    min_nf :
        number of degrees of freedom

    Returns
    -------
        dataframe of categorical cos2
    """
    if model.U is not None and model.s is not None:
        model_coords = pd.DataFrame(
            model.U[:, :min_nf] * model.s[:min_nf],
            columns=get_projected_column_names(min_nf),
        )

    mapping = get_dummies_mapping(model.original_categorical, model.dummy_categorical)
    dummy = pd.get_dummies(
        df[model.original_categorical].astype("category"), prefix_sep=DUMMIES_PREFIX_SEP
    )
    # Compute the categorical cos2 for each original column
    all_category_cos = {}
    for original_col, dummy_columns in mapping.items():
        # FIXME: Kept this for legacy: Why - 1 ? nb_modalities += [nb_dummy_columns - 1] #
        # for each dimension
        all_category_cos[original_col] = _compute_cos2_single_category(
            dummy[dummy_columns], model, model_coords
        )

    # FIXME: Why - 1 ?
    nb_modalities = df[model.original_categorical].nunique() - 1

    categorical_cos2 = pd.DataFrame.from_dict(
        data=all_category_cos,
        orient="index",
        columns=get_projected_column_names(min_nf),
    )

    categorical_cos2 = row_division(categorical_cos2**2, nb_modalities)

    return categorical_cos2


def compute_continuous_cos2(
    model: Model,
    scaled_df: pd.DataFrame,
    min_nf: int,
    s: NDArray[np.float_],
    U: NDArray[np.float_],
) -> pd.DataFrame:
    squared_values: NDArray[np.float_] = scaled_df.values**2

    weighted_values = squared_values * model.row_weights[:, np.newaxis]

    dist2 = np.sum(weighted_values, axis=0)
    dist2 = np.where(np.abs(dist2 - 1) < 0.001, 1, np.sqrt(dist2))

    cos2 = np.divide(U * s, dist2[:, np.newaxis]) ** 2
    # FIXME: Why original_continuous, we have been computing cos2 for all the dummies for nothing?
    continuous_cos2 = pd.DataFrame(
        cos2[: len(model.original_continuous)],  # only keep continuous components
        index=model.original_continuous,
        columns=get_projected_column_names(min_nf),
    )
    continuous_cos2 = continuous_cos2**2
    return continuous_cos2


def _compute_svd(
    model: Model, weighted: NDArray[np.float_], min_nf: int
) -> Tuple[NDArray[np.float_], NDArray[np.float_], NDArray[np.float_]]:
    U, s, V = get_svd(weighted.T, svd_flip=False)

    # Only keep first nf components
    U = U[:, :min_nf]
    V = V.T[:, :min_nf]
    s = s[:min_nf]
    weights = model.column_weights[:min_nf]

    # FIXME: kept this here for legacy, don't know why we were doing it: U, V = V, U

    sign = np.sign(np.sum(U))
    signed_U = column_multiplication(pd.DataFrame(U), sign).values

    # Divide diagonal values by weights
    min_shape = min(signed_U.shape)
    diagonal = np.diag_indices_from(signed_U[:min_shape, :min_shape])
    weighted_U = signed_U
    weighted_U[diagonal] = np.true_divide(signed_U[diagonal], weights)

    eigenvalues = np.power(s, 2)
    return s, weighted_U, eigenvalues


def _compute_cos2_single_category(
    single_category_df: pd.DataFrame, model: Model, coords: pd.DataFrame
) -> NDArray[np.float_]:
    """Compute cos2 for a single original category.

    Parameters
    ----------
    single_category_df:
        dummy dataframe of a single original category
    model :
        _description_
    coords :
        projections of the data used to created the axes

    Returns
    -------
        cos2 of the single category
    """
    cos2 = []
    for coord_col in coords.columns:
        # for each modality of the qualitative column
        p = 0
        for col in single_category_df.columns:
            weighted_coords = coords[coord_col] * model.row_weights
            dummy_values = single_category_df[col].values
            if model.prop is not None:
                p += (dummy_values * weighted_coords).sum() ** 2 / model.prop[col]

        cos2.append(p)
    all_weighted_coords = (coords.values**2).T * model.row_weights
    summed_weights = all_weighted_coords.sum(axis=1)

    single_category_cos2: NDArray[np.float_] = np.array(cos2) / summed_weights
    return single_category_cos2



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['saiph',
 'saiph.lib',
 'saiph.reduction',
 'saiph.reduction.utils',
 'saiph.tests']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.21,<2.0',
 'pandas>=1.3,<2.0',
 'pydantic>=1.10.4,<2.0.0',
 'scikit-learn>=1.0,<2.0',
 'scipy>=1.7,<2.0',
 'toolz>=0.11.2,<0.12.0']

extras_require = \
{'matplotlib': ['matplotlib>=3.5.2,<4.0.0']}

setup_kwargs = {
    'name': 'saiph',
    'version': '1.4.3',
    'description': 'A projection package',
    'long_description': 'None',
    'author': 'Octopize',
    'author_email': 'help@octopize.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<3.12',
}


setup(**setup_kwargs)



from typing import Optional, Tuple

import numpy as np
import pandas as pd
from numpy.typing import NDArray
from scipy import linalg
from sklearn.utils import extmath


def get_svd(
    df: pd.DataFrame,
    nf: Optional[int] = None,
    *,
    svd_flip: bool = True,
    seed: Optional[int] = None,
) -> Tuple[NDArray[np.float_], NDArray[np.float_], NDArray[np.float_]]:
    """Compute Singular Value Decomposition.

    Arguments
    ---------
        df: Matrix to decompose, shape (m, n)
        nf: target number of dimensions to retain (number of singular values). Default `None`.
            It keeps the `nf` higher singular values and nf associated singular vectors.
        svd_flip: Whether to use svd_flip on U and V or not. Default `True`
        seed: random seed. Default `None`

    Returns
    -------
        U: unitary matrix having left singular vectors as columns, shape (m,l)
        S: vector of the singular values, shape (l,)
        Vt: unitary matrix having right singular vectors as rows, shape (l,n)
    """
    if nf is not None and nf < 0.8 * np.min(df.shape):
        # Compute a truncated SVD using randomized methods, for faster computations
        U, S, Vt = get_direct_randomized_svd(
            df, q=2, l_retained_dimensions=nf, seed=seed
        )

    else:
        # Compute a regular full SVD
        U, S, Vt = linalg.svd(df, full_matrices=False)

    if svd_flip:
        U, Vt = extmath.svd_flip(U, Vt)

    return U, S, Vt


def get_randomized_subspace_iteration(
    A: NDArray[np.float_],
    l_retained_dimensions: int,
    *,
    q: int = 2,
    seed: Optional[int] = None,
) -> NDArray[np.float_]:
    """Generate a subspace for more efficient SVD computation using random methods.

    From https://arxiv.org/abs/0909.4061, algorithm 4.4 page 27
    (Finding structure with randomness: Probabilistic algorithms for constructing approximate
    matrix decompositions. Halko, Nathan and Martinsson, Per-Gunnar and Tropp, Joel A.)

    Arguments
    ---------
        A: input matrix, shape (m, n)
        l_retained_dimensions: target number of retained dimensions, l<min(m,n)
        q: exponent of the power method. The higher this exponent, the more precise will be
            the SVD, but more complex to compute. Default `2`
        seed: random seed. Default `None`

    Returns
    -------
        Q: matrix whose range approximates the range of A, shape (m, l)
    """
    m, n = A.shape
    random_gen = np.random.default_rng(seed=seed)
    omega = random_gen.normal(loc=0, scale=1, size=(n, l_retained_dimensions))

    # Initialization
    Y = A @ omega
    Q, _ = np.linalg.qr(Y)

    # Iteration
    for _ in range(q):
        Ytilde = A.transpose() @ Q
        Qtilde, _ = np.linalg.qr(Ytilde)
        Y = A @ Qtilde
        Q, _ = np.linalg.qr(Y)
    return Q


def get_direct_randomized_svd(
    A: NDArray[np.float_],
    l_retained_dimensions: int,
    q: int = 2,
    seed: Optional[int] = None,
) -> Tuple[NDArray[np.float_], NDArray[np.float_], NDArray[np.float_]]:
    """Compute a fixed-rank SVD approximation using random methods.

    The computation of the randomized SVD is generally faster than a regular SVD when we retain
    a smaller number of dimensions than the dimension of the matrix.

    From https://arxiv.org/abs/0909.4061, algorithm 5.1 page 29
    (Finding structure with randomness: Probabilistic algorithms for constructing approximate
    matrix decompositions. Halko, Nathan and Martinsson, Per-Gunnar and Tropp, Joel A.)

    Arguments
    ---------
        A: input matrix, shape (m, n)
        l_retained_dimensions: target number of retained dimensions, l<min(m,n)
        q: exponent of the power method. Higher this exponent, the more precise will be
        the SVD, but more complex to compute.
        seed: random seed. Default `None`

    Returns
    -------
        U: unitary matrix having left singular vectors as columns, shape (m,l)
        S: vector of the singular values, shape (l,)
        Vt: unitary matrix having right singular vectors as rows, shape (l,n)
    """
    is_transposed = False
    if A.shape[1] > A.shape[0]:
        A = A.transpose()
        is_transposed = True

    # Q: matrix whose range approximates the range of A, shape (m, l)
    Q = get_randomized_subspace_iteration(
        A, q=q, l_retained_dimensions=l_retained_dimensions, seed=seed
    )

    B = Q.transpose() @ A

    Utilde, S, Vt = np.linalg.svd(B, full_matrices=False)

    U = Q @ Utilde

    if is_transposed:
        U_bis = U
        Vt_bis = Vt

        U = Vt_bis.transpose()
        Vt = U_bis.transpose()

    return U, S, Vt
